__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging
from multiprocessing import Pool
from multiprocessing.pool import ThreadPool
import sys
from argparse import ArgumentParser, Namespace
from collections import defaultdict
from datetime import timedelta
from functools import partial
from operator import itemgetter
from os import get_terminal_size
from pathlib import Path
from pprint import pformat
from typing import List, Dict

from media_tools.util.length_reader import (
    AudioreadLengthReader, Eyed3LengthReader, MutagenLengthReader, SoundfileLengthReader
)
from media_tools.util.logging import setup_logging

AUDIO_EXTENSIONS = ('mp3', 'ogg', 'flac', 'wav', 'm4a', 'm4b', 'm4p', 'aif', 'aiff', 'ape', 'spx')


def parse_commandline(args: List[str]) -> Namespace:
    parser = ArgumentParser(description='Print length of audio files in directories')
    parser.add_argument('directory', type=str, help='root directory')
    parser.add_argument(
        '-r', '--recursive', action='store_true', help='print run time for every folder'
    )
    parser.add_argument('--debug', action='store_true')
    parser.add_argument('--quiet', action='store_true')
    parser.add_argument('--max-debug-items', type=int, default=-1, help='N shortest items to print')
    return parser.parse_args(args)


class LengthStore:

    def __init__(self, root: Path, recursive: bool):
        self.root = root
        self.recursive = recursive
        self.lengths: Dict[Path, timedelta] = {}
        self.errors: Dict[Path, float] = {}
        self.backends_used: Dict[str, Dict[Path, float]] = defaultdict(dict)

    def scan(self):
        self.lengths[self.root] = self.dir_length(self.root)

    def print_metadata(self, num_entries: int = -1):
        width, _ = get_terminal_size(0)
        logging.debug(pformat({
            backend: [
                f'{track.relative_to(self.root)}: {length:.1f}s'
                for track, length in sorted(tracks.items(), key=itemgetter(1))[:num_entries]
            ]
            for backend, tracks in self.backends_used.items()
        }, width=width))

    def print_errors(self, num_entries: int = -1):
        if not self.errors:
            return
        width, _ = get_terminal_size(0)
        logging.warning('%s errors', len(self.errors))
        logging.warning(pformat([
            f'{track.relative_to(self.root)}: {size / 1024:.1f}kB'
            for track, size in sorted(self.errors.items(), key=itemgetter(1))[:num_entries]
        ], width=width))

    def dir_length(self, folder: Path, parallel: bool = True) -> timedelta:
        audio, folders = get_audio_and_folders(folder)

        length = self.sum_audio_lengths(audio)

        dir_length = partial(self.dir_length, parallel=False)
        if parallel:
            # for recursive printing of run times we need threads, else can use processes
            # using "processes=1" is the experimentally determined ideal number of threads
            with self.create_pool() as pool:
                length = sum(pool.imap_unordered(dir_length, folders), length)
        else:
            length = sum(map(dir_length, folders), length)

        if length:
            self.lengths[folder] = length

        return length

    def create_pool(self):
        return ThreadPool(processes=1) if self.recursive else Pool()

    def sum_audio_lengths(self, paths: List[Path]) -> timedelta:
        return sum([as_timedelta(self.get_duration(path)) for path in paths], timedelta(seconds=0))

    def print(self) -> None:
        max_duration = self.format_timedelta(self.lengths[self.root])
        for path, duration in sorted(self.lengths.items(), key=lambda item: item[1]):
            self.print_line(
                path if path == self.root else path.relative_to(self.root),
                self.format_timedelta(duration).rjust(len(max_duration))
            )

    @staticmethod
    def format_timedelta(duration: timedelta) -> str:
        days = duration.days
        rest = duration - timedelta(days=days)
        hours = str(rest).split('.', maxsplit=1)[0]
        return f'{days}d {hours:>8}' if days else hours

    @staticmethod
    def print_line(path: Path, duration: str) -> None:
        print(f'{duration}   {path}')

    def get_duration(self, track: Path) -> float:
        for reader_class in (
                MutagenLengthReader, Eyed3LengthReader, SoundfileLengthReader, AudioreadLengthReader
        ):
            duration = reader_class(track).get_duration()
            if duration:
                self.backends_used[reader_class.__name__][track] = duration
                return duration
        self.errors[track] = track.stat().st_size
        return 0


def is_audio(track: Path):
    return track.suffix[1:].lower() in AUDIO_EXTENSIONS and track.is_file()


def get_audio_and_folders(folder: Path):
    audio, folders = [], []
    for thing in folder.iterdir():
        if is_audio(thing):
            audio.append(thing)
        elif thing.is_dir():
            folders.append(thing)
    return audio, folders


def as_timedelta(duration: float) -> timedelta:
    try:
        return timedelta(seconds=duration)
    except OverflowError:
        return timedelta(0)


def main() -> None:
    args: List[str] = sys.argv[1:]
    opts = parse_commandline(args)
    setup_logging(opts, fmt='%(message)s')
    lengths = LengthStore(Path(opts.directory), opts.recursive)
    lengths.scan()
    lengths.print()
    if opts.debug:
        lengths.print_errors(opts.max_debug_items)
        lengths.print_metadata(opts.max_debug_items)


if __name__ == '__main__':
    main()



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from functools import wraps
import time
from urllib.error import HTTPError

MAX_TIMES_TRY = 2
WAIT_RETRY_SECS = 0.1


def retry_http(
        max_times_try: int = MAX_TIMES_TRY, silent: bool = True,
        wait_retry_secs: float = WAIT_RETRY_SECS, wait_retry_progressive: bool = False
):

    def actual_decorator(method):
        @wraps(method)
        def allow_fail(self, *args, **kwargs):
            wait_retry = wait_retry_secs
            for i in range(1, max_times_try + 1):
                try:
                    return method(self, *args, **kwargs)
                except HTTPError as error:
                    if not silent:
                        print(
                            f'{type(self).__name__}.{method.__name__}() {error.code}: {error.msg} '
                            f'failed {i} / {max_times_try} times'
                        )
                    if i < max_times_try:
                        time.sleep(wait_retry)
                        if wait_retry_progressive:
                            wait_retry *= 2
            return False

        return allow_fail

    return actual_decorator



#!/usr/bin/env python3

__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging
import os
from argparse import ArgumentParser, Namespace
from pathlib import Path
from shutil import copy2
from sys import argv
from time import time
from typing import List

from media_tools.util import AudaciousTools, find_files
from media_tools.util.logging import setup_logging
from media_tools.util.playlist_copier import PlaylistCopier


def find_newer_than(base_path: str, seconds: int) -> List[Path]:
    return find_files(Path(base_path), lambda file: time() - os.path.getctime(file) < seconds)


def copy_newest_files(src_dir: str, target_dir: str, max_days: int) -> None:
    to_copy = sorted(find_newer_than(src_dir, max_days * 24 * 60 * 60))
    for i, file in enumerate(to_copy):
        basedir = file.parent
        target_subdir = str(basedir).replace(src_dir, '').split(os.path.sep)
        target_path = Path(target_dir).joinpath(*target_subdir)
        target_path.mkdir(exist_ok=True)
        logging.info('%s/%s %s', i + 1, len(to_copy), str(file).replace(src_dir, '').strip('/'))
        if not file.joinpath(target_path).is_file():
            copy_ignoring_errors(file, target_path)


def copy_ignoring_errors(file: Path, target_path: Path) -> None:
    try:
        copy2(file, target_path)
    except OSError:
        pass


def parse_commandline(args: List[str]) -> Namespace:
    parser = ArgumentParser(
        description='Copy the first N existing files of an audacious playlist to a target folder'
    )
    subparsers = parser.add_subparsers(help='Available subcommands:', dest='subparser')
    # workaround for python issue https://bugs.python.org/issue26510
    subparsers.required = True  # type: ignore

    parser_copy = subparsers.add_parser(
        'copy', help='Copy files on the playlist to a specified target folder'
    )
    parser_copy.add_argument(
        '-p', '--playlist', type=str,
        help='ID of the playlist to copy (default: currently playing)'
    )
    parser_copy.add_argument(
        '-n', '--number', default=0, type=int,
        help='First N files to copy from the playlist (default: all)'
    )
    parser_copy.add_argument(
        '-r', '--renumber', action='store_true',
        help='Rename files to have playlist position prepended to file name'
    )
    parser_copy.add_argument('target', type=str, help='Name of the target folder')

    parser_restore = subparsers.add_parser(
        'restore', help='Move files back to the place in the file system they have on the playlist'
    )
    parser_restore.add_argument(
        '-p', '--playlist', type=str,
        help='ID of the playlist to copy (default: currently playing)'
    )

    parser_copy_newest = subparsers.add_parser(
        'copy_newest', help='Copy the latest files to a specified location'
    )
    parser_copy_newest.add_argument(
        '--max-age', type=int, help='Copy files newer than this many days'
    )
    parser_copy_newest.add_argument(
        '-s', '--source', type=str,
        default=os.path.expanduser('~/Music'),
        help='Source directory. Default: ' + os.path.expanduser('~/Music')
    )
    parser_copy_newest.add_argument('target', type=str, help='Name of the target folder')

    parser.add_argument(
        '-v', '--verbose', action='store_true'
    )

    return parser.parse_args(args)


def main() -> None:
    opts = parse_commandline(argv[1:])
    setup_logging(opts)
    audacious = AudaciousTools()
    if opts.subparser == 'copy_newest':
        copy_newest_files(opts.source, opts.target, opts.max_age)
    elif opts.subparser == 'restore':
        copier = PlaylistCopier(audacious, opts.playlist)
        copier.move_files_to_original_places()
    elif opts.subparser == 'copy':
        copier = PlaylistCopier(audacious, opts.playlist)
        copier.copy_playlist(opts.number, Path(opts.target), opts.renumber)
    else:
        logging.critical('what are you doing?')


if __name__ == '__main__':
    main()



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from .authorization_error import AuthorizationError  # noqa: F401
from .constants import (  # noqa: F401
    DEFAULT_CROSSFADE_MS, DEFAULT_MAX_RETRY, DEFAULT_AUDIO_FILE_TYPES
)
from .globals import create_mix, get_access_token  # noqa: F401
from .mix import Mix  # noqa: F401
from .mix_path import MixPath  # noqa: F401
from .multi_mix import MultiMix  # noqa: F401



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import pickle
from datetime import datetime
from pathlib import Path
from shutil import copy2


class TrackDB(set):

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._backup_db()
        with self.bought_db.open('wb') as file:
            pickle.dump(list(self), file)

    def _backup_db(self) -> None:
        if self.bought_db.is_file():
            copy2(
                str(self.bought_db),
                str(self.bought_db.with_suffix(self.suffix_with_appended_timestamp(self.bought_db)))
            )

    @staticmethod
    def suffix_with_appended_timestamp(filepath: Path) -> str:
        return filepath.suffix + f'.{datetime.now().isoformat(timespec="minutes")}'

    def __init__(self, bought_db: Path) -> None:
        super().__init__()
        self.bought_db = bought_db
        try:
            with bought_db.open('rb') as file:
                for track in pickle.load(file):
                    self.add(track)
        except FileNotFoundError:
            pass



from .audacious_tools import AudaciousTools  # noqa: F401
from .base_filename_cleaner import BaseFilenameCleaner  # noqa: F401
from .command import Command, Commands, Move, Nothing  # noqa: F401
from .duplicate_string_remover import DuplicateStringRemover  # noqa: F401
from .junk_remover import JunkRemover  # noqa: F401
from .numbering_fixer import NumberingFixer  # noqa: F401
from .symlink_fixer import SymlinkFixer  # noqa: F401
from .undo_commands import UndoCommands  # noqa: F401
from .util import find_dirs, find_files, find_potential_files, log_utf8_safe  # noqa: F401



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from pathlib import Path


MIXCLOUD_MAX_FILESIZE = 512 * 1024 * 1024
MIXCLOUD_MAX_TAGS = 5
MIXCLOUD_API_UPLOAD_URL = 'https://api.mixcloud.com/upload'
DEFAULT_CROSSFADE_MS = 1000
DEFAULT_MAX_RETRY = 20
DEFAULT_AUDIO_FILE_TYPES = ['flac', 'MP3', 'mp3', 'ogg', 'm4a']
# Mixcloud app:
# https://www.mixcloud.com/developers/Q29uc3VtZXI6MzY3Nw%253D%253D/
ACCESS_TOKEN_FILE = '.mixcloud_access_token'
ACCESS_TOKEN_SEARCH_PATH = (Path('.'), Path.home(), Path.home() / '.config' / 'media-tools')
MP3_KBIT_RATE = 128



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from pathlib import Path
from typing import Optional

from media_tools.util.mixcloud.authorization_error import AuthorizationError
from media_tools.util.mixcloud.constants import (
    MP3_KBIT_RATE, DEFAULT_CROSSFADE_MS, MIXCLOUD_MAX_FILESIZE, ACCESS_TOKEN_FILE,
    ACCESS_TOKEN_SEARCH_PATH
)
from media_tools.util.mixcloud.mix import Mix
from media_tools.util.mixcloud.mix_path import MixPath
from media_tools.util.mixcloud.multi_mix import MultiMix


def bytes_per_second(mp3_kbit_rate: int = MP3_KBIT_RATE) -> int:
    return mp3_kbit_rate // 8 * 1024 * 2


def create_mix(
        mix_path: MixPath, access_token: str, strict: bool = False,
        crossfade_ms: int = DEFAULT_CROSSFADE_MS
) -> 'Mix':
    files, length = mix_path.scan()
    if bytes_per_second() * length < MIXCLOUD_MAX_FILESIZE:
        return Mix(mix_path, strict, crossfade_ms).with_access_token(access_token)
    return MultiMix(
        MixPath(mix_path.basedir, tuple(files)), access_token, length, strict, crossfade_ms
    )


def get_access_token(access_token_path: Optional[Path] = None) -> str:
    error_message = f"""
Authorization token does not exist - please follow the instructions at
https://www.mixcloud.com/developers/#authorization to generate an auth
token and store it under ./{ACCESS_TOKEN_FILE}, ~/{ACCESS_TOKEN_FILE}
or ~/.config/{ACCESS_TOKEN_FILE}.
"""
    if not access_token_path:
        access_token_path = find_access_token(error_message)
    try:
        with access_token_path.open('r') as file:
            return file.read().strip()
    except OSError as error:
        raise AuthorizationError(error_message) from error


def find_access_token(error_message):
    for basedir in ACCESS_TOKEN_SEARCH_PATH:
        if (basedir / ACCESS_TOKEN_FILE).exists():
            return basedir / ACCESS_TOKEN_FILE
    raise AuthorizationError(error_message)



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging
from math import ceil
from pathlib import Path
from typing import List

from media_tools.util.mixcloud.constants import (
    DEFAULT_CROSSFADE_MS, MIXCLOUD_MAX_FILESIZE, DEFAULT_MAX_RETRY, MP3_KBIT_RATE
)
from media_tools.util.mixcloud.mix import Mix
from media_tools.util.mixcloud.mix_path import MixPath


def bytes_per_second(mp3_kbit_rate: int = MP3_KBIT_RATE) -> int:
    return mp3_kbit_rate // 8 * 1024 * 2


# noinspection PyMissingConstructor
class MultiMix(Mix):

    def __init__(  # pylint: disable=super-init-not-called, too-many-arguments
            self, mix_path: MixPath, access_token: str, total_length: int,
            strict: bool = False, crossfade_ms: int = DEFAULT_CROSSFADE_MS
    ) -> None:
        self._mix_path = mix_path
        self._mix_parts: List[Mix] = []
        self._part_paths: List[Path] = []
        self._incomplete = False
        oversize_factor = ceil(total_length * bytes_per_second() / MIXCLOUD_MAX_FILESIZE)
        chunk_size = len(self._mix_path.audio_files) // oversize_factor
        for i in range(oversize_factor):
            part_files = tuple(self._mix_path.audio_files[i * chunk_size:(i + 1) * chunk_size])
            part_name = f'{self.title} Part {i + 1}'
            logging.info('MultiMix: %s', part_name)
            mix_part = Mix(
                self._mix_path.from_files(part_files),
                strict=strict, crossfade_ms=crossfade_ms, part=i + 1
            ).with_access_token(access_token).with_title(part_name)
            self._mix_parts.append(mix_part)
        self._incomplete = any(part._incomplete for part in self._mix_parts)

    @property
    def parts(self):
        return self._mix_parts

    @property
    def title(self) -> str:
        title = self._mix_path.title
        return f"Test - don't bother playing ({title})" if self._incomplete else title

    def upload(self, name: Path = Path('mix.mp3'), max_retry: int = DEFAULT_MAX_RETRY) -> None:
        for mix_part, part_path in zip(self._mix_parts, self._part_paths):
            mix_part.upload(part_path, max_retry)

    def export(self, name: Path = Path('mix.mp3')) -> None:
        for i, mix_part in enumerate(self._mix_parts):
            part_path = Path(f'{name.stem}_{i + 1}{name.suffix}')
            mix_part.export(part_path)
            self._part_paths.append(part_path)



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging
import sys
from pathlib import Path
from time import sleep
from typing import List, Dict, Optional

import requests
from pydub import AudioSegment
from tinytag import TinyTag

from media_tools.util.mixcloud.constants import (
    DEFAULT_CROSSFADE_MS, DEFAULT_MAX_RETRY, MP3_KBIT_RATE, MIXCLOUD_API_UPLOAD_URL,
    MIXCLOUD_MAX_TAGS
)
from media_tools.util.mixcloud.mix_path import MixPath


class Mix:  # pylint: disable=too-many-instance-attributes

    def __init__(
            self, mix_path: MixPath, strict: bool = False, crossfade_ms: int = DEFAULT_CROSSFADE_MS,
            part: Optional[int] = None
    ) -> None:
        self._mix_path = mix_path
        self._strict = strict
        self._crossfade_ms = crossfade_ms
        self._incomplete = False
        self._track_info: List[Dict] = []
        self._title: Optional[str] = None
        self._access_token: Optional[str] = None
        self._part: Optional[int] = part
        if self._strict and not self._mix_path.audio_files:
            raise ValueError(f'no audio files in {self._mix_path.basedir}')
        self._audio = self._import_audio(self._mix_path.audio_files)

    def with_access_token(self, access_token: str) -> 'Mix':
        self._access_token = access_token
        return self

    def with_title(self, title: str) -> 'Mix':
        self._title = title
        return self

    @property
    def valid(self):
        return not self._incomplete

    @property
    def tags(self) -> List[str]:
        tags = self._mix_path.tags
        error = self.check_tag_validity(tags)
        if error:
            self._incomplete = True
            logging.warning(str(error))
            if self._strict:
                raise ValueError(error)
        return tags

    @staticmethod
    def check_tag_validity(tags: List[str]) -> str:
        if len(tags) > MIXCLOUD_MAX_TAGS:
            return f'Max. {MIXCLOUD_MAX_TAGS} tags allowed, found {len(tags)}: {tags}'
        if not tags:
            return 'No tags found'
        return ''

    @property
    def title(self) -> str:
        if self._title:
            return self._title
        title = self._mix_path.title
        return f"Test - don't bother playing ({title})" if self._incomplete else title

    @property
    def description(self) -> str:
        description = self._mix_path.description
        if description:
            return description
        if self._strict:
            raise ValueError('No description found')
        logging.warning('No description found')
        self._incomplete = True
        return ''

    @property
    def picture(self) -> Optional[Path]:
        """Currently just returns the first JPEG or PNG. Room for improvement!"""
        try:
            return self._mix_path.picture
        except StopIteration as error:
            if self._strict:
                raise ValueError(f'No picture in {self._mix_path.basedir}') from error
            logging.warning('No picture found')
            self._incomplete = True
            return None

    def _import_audio(self, audio_files: List[Path]) -> AudioSegment:
        part_name = '' if self._part is None else f'_{self._part}'
        mix_file = self._mix_path.file(Path(f'mix{part_name}.mp3'))
        self._track_info = [self._get_track_info(audio_file) for audio_file in audio_files]
        if mix_file.exists():
            return AudioSegment.from_file(mix_file)

        audio = AudioSegment.empty()
        for i, audio_file in enumerate(audio_files):
            logging.info('%s - %s', i + 1, audio_file.name)
            audio = audio.append(
                self._read_track(audio_file),
                crossfade=self._crossfade_ms if len(audio) > self._crossfade_ms else len(audio)
            )

        return audio

    @staticmethod
    def _read_track(audio_file):
        track = AudioSegment.from_file(audio_file)
        if track.sample_width != 2:
            track = track.set_sample_width(2)
        if track.frame_rate != 44100:
            track = track.set_frame_rate(44100)
        return track.normalize()

    def _get_track_info(self, audio_file: Path) -> Dict:
        tags = TinyTag.get(str(audio_file))
        if tags.artist is None or tags.title is None:
            if self._strict:
                raise ValueError(f'Incomplete tags for {audio_file}')
            logging.warning('Incomplete tags for %s', audio_file)
            self._incomplete = True
            return {
                'artist': tags.artist or '???', 'title': tags.title or '???',
                'length': tags.duration
            }
        return {
            'artist': tags.artist, 'title': tags.title, 'length': tags.duration,
            'filename': audio_file.name
        }

    def export(self, name: Path = Path('mix.mp3')) -> None:
        mix_file = self._mix_path.file(name)
        if mix_file.exists():
            return
        audio_format = name.suffix[1:]
        logging.info(
            'Exporting %s audio to %s with bitrate %s kbps',
            self.mix_length(), mix_file, MP3_KBIT_RATE
        )
        self._audio.export(
            mix_file, format=audio_format, parameters=['-q:a', '0'], bitrate=f'{MP3_KBIT_RATE}k'
        )

    def upload(self, name: Path = Path('mix.mp3'), max_retry: int = DEFAULT_MAX_RETRY) -> None:
        mix_file = self._mix_path.file(name)
        if not mix_file.exists():
            raise FileNotFoundError(mix_file)
        files = {
            'mp3': ('mix.mp3', mix_file.open('rb'), 'audio/mpeg'),
        }
        if self.picture:
            picture_type = self.picture.suffix
            files['picture'] = (
                'picture' + picture_type, self.picture.open('rb'),
                f'image/{"png" if picture_type == ".png" else "jpeg"}'
            )

        data = {
            'name': self.title,
            'description': self.description,
            'percentage_music': 100
        }
        self._add_tags(data)
        self._add_track_info(data)
        size_bytes = mix_file.stat().st_size + self.picture.stat().st_size if self.picture else 0
        logging.info(
            'Uploading %s kBytes (%s) as %s',
            f'{size_bytes // 1024:,d}', self.mix_length(), self.title
        )
        self._do_upload(name, files, data, max_retry)

    def mix_length(self) -> str:
        return f'{len(self._audio) // 60000}:{len(self._audio) % 60000 // 1000:02}'

    def _do_upload(self, name: Path, files: Dict, data: Dict, max_retry: int):
        if max_retry < 0:
            logging.error('out of retries, aborting upload attempt')
            sys.exit(1)

        url = MIXCLOUD_API_UPLOAD_URL + '/?access_token=' + self._access_token
        logging.info('uploading to %s, max. %s retries', url, max_retry)
        try:
            response = requests.post(url, files=files, data=data)
        except requests.exceptions.ConnectionError as error:
            logging.warning(self.connection_error_message(error, max_retry))
            self._do_upload(name, files=files, data=data, max_retry=max_retry - 1)
        else:
            if self.is_rate_limited(response):
                logging.info(self.response_message(response))
                sleep(int(response.json().get('error', {}).get('retry_after', 60)))
                self._do_upload(name, files=files, data=data, max_retry=max_retry - 1)
            elif response.status_code == 200:
                logging.info(self.response_message(response))
                sys.exit(0)
            else:
                logging.error(self.response_message(response))
                sys.exit(1)

    @staticmethod
    def is_rate_limited(response: requests.Response) -> bool:
        return (response.status_code == 403 and
                response.json().get('error', {}).get('type') == 'RateLimitException') or (
            response.status_code == 400 and
            response.json().get('error', {}).get('type') == 'PostValidationError'
        )

    @staticmethod
    def response_message(response: requests.Response) -> str:
        if response.status_code == 200:
            return f"{response.status_code}: {response.json()['result']['message']}"
        return f"{response.status_code}: {response.json()['error']}"

    @staticmethod
    def connection_error_message(error: requests.exceptions.ConnectionError, max_retry: int) -> str:
        message = Mix.error_args_message(error)
        if max_retry > 0:
            return f"{message} Retrying {max_retry} time{'s' if max_retry > 1 else ''}."
        return f'{message} Giving up.'

    @staticmethod
    def error_args_message(error: requests.exceptions.ConnectionError) -> str:
        if not error.args:
            return ''
        if error.args[0].args and len(error.args[0].args) > 1:
            return f'{error.args[0].args[1].args[1]}: {error.args[0].args[0]} '
        return str(error.args)

    def _add_tags(self, data: Dict) -> None:
        for i, tag in enumerate(self.tags):
            data[f'tags-{i}-tag'] = tag

    def _add_track_info(self, data: Dict) -> None:
        start_time = 0
        for i, track_info in enumerate(self._track_info):
            data[f'sections-{i}-artist'] = track_info['artist']
            data[f'sections-{i}-song'] = track_info['title']
            data[f'sections-{i}-start_time'] = int(start_time)
            start_time += (track_info['length'] - self._crossfade_ms / 1000)



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import json
import pickle
from bz2 import open as bzopen
from gzip import open as gzopen
from lzma import open as xzopen
from typing import Callable, Dict

OPEN_FUNCTIONS: Dict[str, Callable] = {
    'bz2': bzopen,
    'gz': gzopen,
    'xz': xzopen,
}
DUMP_PARAMS: Dict[str, Dict] = {
    'json': {'dump': json.dump, 'mode': 'wt', 'kwargs': dict(default=str)},
    'pickle': {'dump': pickle.dump, 'mode': 'wb', 'kwargs': {}}
}


def open_function(extension: str) -> Callable:
    return OPEN_FUNCTIONS.get(extension, open)


def dump_data(user_data: Dict, backup_filename: str) -> None:
    filename_parts = backup_filename.split('.')
    extension = filename_parts[-1].lower()
    dump_format = extension if extension in DUMP_PARAMS else filename_parts[-2].lower()
    dump_params = DUMP_PARAMS.get(dump_format, DUMP_PARAMS['pickle'])
    dump_func = dump_params['dump']
    mode = dump_params['mode']
    kwargs = dump_params['kwargs']
    with open_function(extension)(backup_filename, mode) as backup_file:
        dump_func(user_data, backup_file, **kwargs)
